"""A module containing the bindings for the histogram class.

This module contains the function that defines the bindings for the histogram
and attaches them to the application. It should not be used directly.
"""

from prompt_toolkit.document import Document
from prompt_toolkit.filters import Condition
from prompt_toolkit.layout import ConditionalContainer, VSplit
from prompt_toolkit.widgets import Label

from h5forest.errors import error_handler


def _init_hist_bindings(app):
    """Set up the bindings for histogram mode."""

    @error_handler
    def edit_hist_entry(event):
        """Edit histogram param under cursor."""
        # Get the current position and row in the plot content
        current_row = app.hist_content.document.cursor_position_row
        current_pos = app.hist_content.document.cursor_position

        # Get the current row text  in the plot content split into
        # key and value
        split_line = app.histogram_plotter.get_row(current_row).split(": ")

        # Split the current plot content into lines
        split_text = app.hist_content.text.split("\n")

        # If we're on a toggle option (i.e. scaling is linear or log) lets
        # toggle it rather than edit it
        if "scale" in split_line[0]:
            if split_line[1].strip() == "linear":
                split_text[current_row] = (
                    f"{split_line[0]}:  ".ljust(13) + "log"
                )
            else:
                split_text[current_row] = (
                    f"{split_line[0]}:  ".ljust(13) + "linear"
                )

            app.hist_content.text = "\n".join(split_text)

            # And put the cursor back where it was
            app.hist_content.document = Document(
                text=app.hist_content.text, cursor_position=current_pos
            )
            app.histogram_plotter.plot_text = app.hist_content.text

            app.app.invalidate()

            return

        def edit_hist_entry_callback():
            """Edit the plot param under cursor."""
            # Strip the user input
            user_input = app.user_input.strip()

            # And set the text here
            split_text[current_row] = (
                f"{split_line[0]}:  ".ljust(13) + f"{user_input}"
            )

            # And display the new text
            app.hist_content.text = "\n".join(split_text)
            app.histogram_plotter.plot_text = app.hist_content.text

            # And shift focus back to the plot content
            app.shift_focus(app.hist_content)

            # And put the cursor back where it was
            app.hist_content.document = Document(
                text=app.hist_content.text, cursor_position=current_pos
            )

        # Get the modified entry from the user
        app.input(split_line[0], edit_hist_entry_callback)

    @error_handler
    def plot_hist(event):
        """Plot the histogram."""
        # Don't update if we already have everything
        if len(app.histogram_plotter.plot_params) == 0:
            # Get the node under the cursor
            node = app.tree.get_current_node(app.current_row)

            # Exit if the node is not a Dataset
            if node.is_group:
                app.print(f"{node.path} is not a Dataset")
                return

            # Set the text in the plotting area
            app.hist_content.text = app.histogram_plotter.set_data_key(node)

        # Compute the histogram
        app.hist_content.text = app.histogram_plotter.compute_hist(
            app.hist_content.text
        )

        # Get the plot
        app.histogram_plotter.plot_and_show(app.hist_content.text)

    @error_handler
    def save_hist(event):
        """Plot the histogram."""
        # Don't update if we already have everything
        if len(app.histogram_plotter.plot_params) == 0:
            # Get the node under the cursor
            node = app.tree.get_current_node(app.current_row)

            # Exit if the node is not a Dataset
            if node.is_group:
                app.print(f"{node.path} is not a Dataset")
                return

            # Set the text in the plotting area
            app.hist_content.text = app.histogram_plotter.set_data_key(node)

        # Compute the histogram
        app.hist_content.text = app.histogram_plotter.compute_hist(
            app.hist_content.text
        )

        # Get the plot
        app.histogram_plotter.plot_and_save(app.hist_content.text)

    @error_handler
    def reset_hist(event):
        """Reset the histogram content."""
        app.hist_content.text = app.histogram_plotter.reset()
        app.return_to_normal_mode()
        app.default_focus()

    @error_handler
    def edit_hist(event):
        """Edit the histogram."""
        app.shift_focus(app.hist_content)

    def exit_edit_hist(event):
        """Exit the edit mode."""
        app.shift_focus(app.tree_content)

    # Bind the functions
    app.kb.add(
        "enter",
        filter=Condition(lambda: app.app.layout.has_focus(app.hist_content)),
    )(edit_hist_entry)
    app.kb.add("h", filter=Condition(lambda: app.flag_hist_mode))(plot_hist)
    app.kb.add("H", filter=Condition(lambda: app.flag_hist_mode))(save_hist)
    app.kb.add("r", filter=Condition(lambda: app.flag_hist_mode))(reset_hist)
    app.kb.add(
        "e",
        filter=Condition(
            lambda: app.flag_hist_mode
            and len(app.histogram_plotter.plot_params) > 0
        ),
    )(edit_hist)
    app.kb.add(
        "q",
        filter=Condition(lambda: app.app.layout.has_focus(app.hist_content)),
    )(exit_edit_hist)

    # Add the hot keys
    hot_keys = VSplit(
        [
            ConditionalContainer(
                Label("e → Edit Config"),
                Condition(lambda: len(app.histogram_plotter.plot_params) > 0),
            ),
            ConditionalContainer(
                Label("Enter → Edit entry"),
                Condition(lambda: app.app.layout.has_focus(app.hist_content)),
            ),
            Label("h → Show Histogram"),
            Label("H → Save Histogram"),
            Label("r → Reset"),
            ConditionalContainer(
                Label("q → Exit Hist Mode"),
                Condition(
                    lambda: not app.app.layout.has_focus(app.hist_content)
                ),
            ),
            ConditionalContainer(
                Label("q → Exit Config"),
                Condition(lambda: app.app.layout.has_focus(app.hist_content)),
            ),
        ]
    )

    return hot_keys



"""A module containing the keybindings for the basic UI.

This module contains the keybindings for the basic UI. These keybindings are
always active and are not dependent on any leader key. The functions in this
module should not be called directly, but are intended to be used by the main
application.
"""

from prompt_toolkit.filters import Condition
from prompt_toolkit.layout import ConditionalContainer
from prompt_toolkit.widgets import Label

from h5forest.errors import error_handler


def _init_app_bindings(app):
    """
    Set up the keybindings for the basic UI.

    This includes basic closing functionality and leader keys for different
    modes. These are always active and are not dependent on any leader key.
    """

    def exit_app(event):
        """Exit the app."""
        event.app.exit()

    def goto_leader_mode(event):
        """Enter goto mode."""
        app._flag_normal_mode = False
        app._flag_jump_mode = True

    def dataset_leader_mode(event):
        """Enter dataset mode."""
        app._flag_normal_mode = False
        app._flag_dataset_mode = True

    def window_leader_mode(event):
        """Enter window mode."""
        app._flag_normal_mode = False
        app._flag_window_mode = True

    def plotting_leader_mode(event):
        """Enter plotting mode."""
        app._flag_normal_mode = False
        app._flag_plotting_mode = True

    def hist_leader_mode(event):
        """Enter hist mode."""
        app._flag_normal_mode = False
        app._flag_hist_mode = True

    @error_handler
    def exit_leader_mode(event):
        """Exit leader mode."""
        app.return_to_normal_mode()
        app.default_focus()
        event.app.invalidate()

    def expand_attributes(event):
        """Expand the attributes."""
        app.flag_expanded_attrs = True
        event.app.invalidate()

    def collapse_attributes(event):
        """Collapse the attributes."""
        app.flag_expanded_attrs = False
        event.app.invalidate()

    # Bind the functions
    app.kb.add("q", filter=Condition(lambda: app.flag_normal_mode))(exit_app)
    app.kb.add("c-q")(exit_app)
    app.kb.add("g", filter=Condition(lambda: app.flag_normal_mode))(
        goto_leader_mode
    )
    app.kb.add("d", filter=Condition(lambda: app.flag_normal_mode))(
        dataset_leader_mode
    )
    app.kb.add("w", filter=Condition(lambda: app.flag_normal_mode))(
        window_leader_mode
    )
    app.kb.add("p", filter=Condition(lambda: app.flag_normal_mode))(
        plotting_leader_mode
    )
    app.kb.add("H", filter=Condition(lambda: app.flag_normal_mode))(
        hist_leader_mode
    )
    app.kb.add("q", filter=Condition(lambda: not app.flag_normal_mode))(
        exit_leader_mode
    )
    app.kb.add(
        "A",
        filter=Condition(
            lambda: app.flag_normal_mode and not app.flag_expanded_attrs
        ),
    )(expand_attributes)
    app.kb.add(
        "A",
        filter=Condition(
            lambda: app.flag_normal_mode and app.flag_expanded_attrs
        ),
    )(collapse_attributes)

    # Add the hot keys
    hot_keys = [
        ConditionalContainer(
            Label("A → Expand Attributes"),
            filter=Condition(lambda: not app.flag_expanded_attrs),
        ),
        ConditionalContainer(
            Label("A → Shrink Attributes"),
            filter=Condition(lambda: app.flag_expanded_attrs),
        ),
        Label("d → Dataset Mode"),
        Label("g → Goto Mode"),
        Label("H → Histogram Mode"),
        Label("p → Plotting Mode"),
        Label("w → Window Mode"),
        Label("q → Exit"),
    ]

    return hot_keys



"""A module containing functions for graceful error handling."""


def error_handler(func):
    """
    Wrap a function in a try/except block to catch errors.

    Errors are printed to the mini buffer.

    Args:
        func (function):
            The function to wrap.
    """

    def wrapper(*args, **kwargs):
        """Wrap the function."""
        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt:
            # Re-raise the KeyboardInterrupt to ensure it's not caught here
            raise
        except Exception as e:
            # Nested import to avoid circular dependencies
            from h5forest.h5_forest import H5Forest

            H5Forest().print(f"ERROR@{func.__name__}: {e}")

    return wrapper



"""Tree class for the HDF5 file viewer.

This module contains the Tree class which is used to represent the HDF5 file
as a tree structure. The Tree contains Nodes which are used to represent the
groups and datasets in the HDF5 file. Each Node is lazy loaded, meaning that
its children are only loaded when it is opened.

Example usage:
    tree = Tree("example.h5")
    print(tree.get_tree_text())
"""

import h5py
from prompt_toolkit.layout.processors import Processor, Transformation

from h5forest.node import Node


class TreeProcessor(Processor):
    def __init__(self, tree):
        self.tree = tree

    def apply_transformation(self, ti):
        lineno = ti.lineno
        fragments = ti.fragments

        # Access the node corresponding to the current line
        if lineno < len(self.tree.nodes_by_row):
            node = self.tree.nodes_by_row[lineno]
            style = ""

            if node.is_group:
                style += " class:group"
            if node.is_highlighted:
                style += " class:highlighted"
            if node.is_under_cursor:
                style += " class:under_cursor"

            # Apply the style to the entire line
            new_fragments = [(style, text) for _style, text in fragments]
            return Transformation(new_fragments)
        else:
            # Line number exceeds the number of nodes, return original
            # fragments
            return Transformation(fragments)


class Tree:
    """A class to represent the HDF5 file as a tree structure.

    This class is used to represent the HDF5 file as a tree structure. The
    tree is constructed from Node objects representing Groups and Datasets
    which are lazy loaded, meaning that the children of a node are only
    loaded when needed.

    Attributes:
        filepath (str):
            The path to the HDF5 file.
        roots (list):
            A list of the root level nodes in the tree.
        nodes_by_row (list):
            A list of all nodes in the tree by row in the tree output.
    """

    def __init__(self, filepath):
        """
        Initialise the tree.

        This will set up the attributes we'll need and parses the root level of
        the HDF5 file ready to be displayed.

        Args:
            filepath (str):
                The path to the HDF5 file.
        """
        # Store the file path we're working with
        self.filepath = filepath
        self.filename = (
            filepath.split("/")[-1].replace(".h5", "").replace(".hdf5", "")
        )

        # Initialise a container to store nodes by row in the tree output
        self.nodes_by_row = []

        # Intialise containers to hold the tree text and split version
        # to avoid wasted computation
        self.tree_text = ""
        self.tree_text_split = []

        # Get the root of the level
        with h5py.File(self.filepath, "r") as hdf:
            self.root = Node(self.filename, hdf, self.filepath)
            self.root.is_under_cursor = True

        # Store the previous node under the cursor (we need some memory for
        # correct highlighting)
        self.prev_node = self.root

    @property
    def length(self):
        """Return the length of the tree text."""
        return len(self.tree_text)

    @property
    def height(self):
        """Return the height of the tree text."""
        return len(self.nodes_by_row)

    @property
    def width(self):
        """
        Return the width of the tree text.

        Note that this works because every line is padded with spaces to
        the same length.
        """
        return len(self.tree_text_split[0])

    def parse_level(self, parent):
        """
        Open the parent group.

        This will populate the chidlren dict on the node which will be parsed
        later when a text representation is requested updating the tree.

        Args:
            parent (Node):
                The parent node to open.
        """
        # Open this group
        parent.open_node()

    def _get_tree_text_recursive(self, current_node, text, nodes_by_row):
        """
        Parse the open nodes to produce the text tree representation.

        This will recurse through the open nodes constructing the output.

        Args:
            current_node (Node):
                The current node to parse.
            text (str):
                The current text representation of the tree.
            nodes_by_row (list):
                A list containing the nodes where the index is the row
                they are on in the text representation.

        Returns:
            str:
                The text representation of the tree.
            list:
                A list containing the nodes where the index is the row
                they are on in the text representation.
        """
        # Add this nodes representation
        text += f"{current_node.to_tree_text()}\n"

        # Append this node to the by row list
        nodes_by_row.append(current_node)

        # And include any children
        for child in current_node.children:
            text, nodes_by_row = self._get_tree_text_recursive(
                child,
                text,
                nodes_by_row,
            )

        return text, nodes_by_row

    def get_tree_text(self):
        """
        Return the text representation of the tree.

        Note that this is only called at initialisation and thus only parses
        the roots of the tree. Any future updates will be done via
        update_tree_text. This is to avoid recalculating the full tree for
        every change.

        Returns:
            str:
                The text representation of the tree.
        """
        text = ""
        nodes_by_row = []
        text, nodes_by_row = self._get_tree_text_recursive(
            self.root,
            text,
            nodes_by_row,
        )

        # Store the nodes by row
        self.nodes_by_row = nodes_by_row

        # Store the tree text
        self.tree_text = text
        self.tree_text_split = text.split("\n")

        return text

    def update_tree_text(self, parent, current_row):
        """
        Update the tree text for the parent node.

        Args:
            parent (Node):
                The parent node to update.
            current_row (int):
                The row in the tree text where the parent is.
        """
        # Open the parent
        self.parse_level(parent)

        # Update the parent node to reflect that it is now open
        self.tree_text_split[current_row] = parent.to_tree_text()

        # Create the text and node list for the children ready to insert
        child_test = [child.to_tree_text() for child in parent.children]
        child_nodes_by_row = [child for child in parent.children]

        # Insert the children into the tree text and nodes by row list
        self.tree_text_split[current_row + 1 : current_row + 1] = child_test
        self.nodes_by_row[current_row + 1 : current_row + 1] = (
            child_nodes_by_row
        )

        # Update the tree text area
        self.tree_text = "\n".join(self.tree_text_split)

        return self.tree_text

    def close_node(self, node, current_row):
        """
        Close the node.

        Args:
            node (Node):
                The node to close.
            current_row (int):
                The row in the tree text where the node is.
        """
        # Close the node itself
        node.close_node()

        # Now we need to remove all the children from the tree, these have
        # already been closed recursively by the call to `close_node` above
        # so we just need to remove them from the tree text and the nodes by
        # row list

        # We can do this by removing everything between the node and the next
        # node at the same depth
        for i, n in enumerate(self.nodes_by_row[current_row + 1 :]):
            if n.depth <= node.depth:
                break
            del self.nodes_by_row[current_row + 1]
            del self.tree_text_split[current_row + 1]

        # Update the parent node to reflect that it is now closed
        self.tree_text_split[current_row] = node.to_tree_text()

        # Update the tree text area
        self.tree_text = "\n".join(self.tree_text_split)

        return self.tree_text

    def get_current_node(self, row):
        """
        Return the current node.

        This will also unhighlight the previous node, and highlight the new
        node.

        Args:
            row (int):
                The row in the tree.

        Returns:
            Node:
                The node at row.
        """
        # Unhighlight the previous node
        self.prev_node.is_under_cursor = False

        # Get the new node and highlight it
        new_node = self.nodes_by_row[row]
        new_node.is_under_cursor = True

        # New node will now be the previous node
        self.prev_node = new_node

        return new_node



"""This module contains the keybindings for the plotting mode.

The functions in this module are used to define the keybindings for the
plotting mode and attach them to the application. It should not be used
directly.
"""

from prompt_toolkit.document import Document
from prompt_toolkit.filters import Condition
from prompt_toolkit.layout import ConditionalContainer, VSplit
from prompt_toolkit.widgets import Label

from h5forest.errors import error_handler


def _init_plot_bindings(app):
    """Set up the keybindings for the plotting mode."""

    @error_handler
    def select_x(event):
        """Select the x-axis."""
        # Get the node under the cursor
        node = app.tree.get_current_node(app.current_row)

        # Exit if the node is not a Dataset
        if node.is_group:
            app.print(f"{node.path} is not a Dataset")
            return

        # Set the text in the plotting area
        app.plot_content.text = app.scatter_plotter.set_x_key(node)

    @error_handler
    def select_y(event):
        """Select the y-axis."""
        # Get the node under the cursor
        node = app.tree.get_current_node(app.current_row)

        # Exit if the node is not a Dataset
        if node.is_group:
            app.print(f"{node.path} is not a Dataset")
            return

        # Set the text in the plotting area
        app.plot_content.text = app.scatter_plotter.set_y_key(node)

    @error_handler
    def edit_plot_entry(event):
        """Edit plot param under cursor."""
        # Get the current position and row in the plot content
        current_row = app.plot_content.document.cursor_position_row
        current_pos = app.plot_content.document.cursor_position

        # Get the current row text  in the plot content split into
        # key and value
        split_line = app.scatter_plotter.get_row(current_row).split(": ")

        # Split the current plot content into lines
        split_text = app.plot_content.text.split("\n")

        # If we're on a toggle option (i.e. scaling is linear or log) lets
        # toggle it rather than edit it
        if "scale" in split_line[0]:
            if split_line[1].strip() == "linear":
                split_text[current_row] = (
                    f"{split_line[0]}:  ".ljust(13) + "log"
                )
            else:
                split_text[current_row] = (
                    f"{split_line[0]}:  ".ljust(13) + "linear"
                )

            app.plot_content.text = "\n".join(split_text)

            # And put the cursor back where it was
            app.plot_content.document = Document(
                text=app.plot_content.text, cursor_position=current_pos
            )
            app.scatter_plotter.plot_text = app.plot_content.text

            app.app.invalidate()

            return

        def edit_plot_entry_callback():
            """Edit the plot param under cursor."""
            # Strip the user input
            user_input = app.user_input.strip()

            # And set the text here
            split_text[current_row] = (
                f"{split_line[0]}:  ".ljust(13) + f"{user_input}"
            )

            # And display the new text
            app.plot_content.text = "\n".join(split_text)
            app.scatter_plotter.plot_text = app.plot_content.text

            # And shift focus back to the plot content
            app.shift_focus(app.plot_content)

            # And put the cursor back where it was
            app.plot_content.document = Document(
                text=app.plot_content.text, cursor_position=current_pos
            )

        # Get the modified entry from the user
        app.input(split_line[0], edit_plot_entry_callback)

    @error_handler
    def plot_scatter(event):
        """Plot and show pcolormesh with mean in bins."""
        # Make the plot
        app.scatter_plotter.plot_and_show(app.plot_content.text)

        app.return_to_normal_mode()
        app.default_focus()

    @error_handler
    def save_scatter(event):
        """Save the plot."""
        app.scatter_plotter.plot_and_save(app.plot_content.text)

    @error_handler
    def reset(event):
        """Reset the plot content."""
        app.plot_content.text = app.scatter_plotter.reset()

        app.app.invalidate()
        app.return_to_normal_mode()
        app.default_focus()

    @error_handler
    def edit_plot(event):
        """Edit the plot."""
        app.shift_focus(app.plot_content)

    def exit_edit_plot(event):
        """Exit edit plot mode."""
        app.shift_focus(app.tree_content)

    # Bind the functions
    app.kb.add("x", filter=Condition(lambda: app.flag_plotting_mode))(select_x)
    app.kb.add("y", filter=Condition(lambda: app.flag_plotting_mode))(select_y)
    app.kb.add(
        "enter",
        filter=Condition(lambda: app.app.layout.has_focus(app.plot_content)),
    )(edit_plot_entry)
    app.kb.add("p", filter=Condition(lambda: app.flag_plotting_mode))(
        plot_scatter
    )
    app.kb.add("P", filter=Condition(lambda: app.flag_plotting_mode))(
        save_scatter
    )
    app.kb.add("r", filter=Condition(lambda: app.flag_plotting_mode))(reset)
    app.kb.add(
        "e",
        filter=Condition(
            lambda: app.flag_plotting_mode
            and len(app.scatter_plotter.plot_params) > 0
        ),
    )(edit_plot)
    app.kb.add(
        "q",
        filter=Condition(lambda: app.app.layout.has_focus(app.plot_content)),
    )(exit_edit_plot)

    # Add the hot keys
    hot_keys = VSplit(
        [
            ConditionalContainer(
                Label("e → Edit Config"),
                Condition(lambda: len(app.scatter_plotter.plot_params) > 0),
            ),
            ConditionalContainer(
                Label("Enter → Edit entry"),
                Condition(lambda: app.app.layout.has_focus(app.plot_content)),
            ),
            ConditionalContainer(
                Label("x → Select x-axis"),
                filter=Condition(
                    lambda: "x" not in app.scatter_plotter.plot_params
                ),
            ),
            ConditionalContainer(
                Label("y → Select y-axis"),
                filter=Condition(
                    lambda: "y" not in app.scatter_plotter.plot_params
                ),
            ),
            ConditionalContainer(
                Label("p → Plot"),
                Condition(lambda: len(app.scatter_plotter) > 0),
            ),
            ConditionalContainer(
                Label("P → Save Plot"),
                Condition(lambda: len(app.scatter_plotter) > 0),
            ),
            Label("r → Reset"),
            ConditionalContainer(
                Label("q → Exit Plotting Mode"),
                Condition(
                    lambda: not app.app.layout.has_focus(app.plot_content)
                ),
            ),
            ConditionalContainer(
                Label("q → Exit Config"),
                Condition(lambda: app.app.layout.has_focus(app.plot_content)),
            ),
        ]
    )

    return hot_keys



"""A module containing utility functions and classes for the HDF5 viewer."""

import os


class DynamicTitle:
    """
    A class to represent a dynamic title for the application.

    This can be used to update any title in the application dynamically.

    Attributes:
         title (str): The title to display.
    """

    def __init__(self, initial_title="Initial Title"):
        """
        Initialise the title.

        Args:
            initial_title (str): The initial title to display.
        """
        self.title = initial_title

    def __call__(self):
        """Return the current title."""
        return self.title

    def update_title(self, new_title):
        """
        Update the title.

        Args:
            new_title (str): The new title to display.
        """
        self.title = new_title


def get_window_size():
    """
    Get the terminal window size in lines and characters.

    Returns:
        tuple: The number of lines and characters in the terminal window.
    """
    rows, columns = os.popen("stty size", "r").read().split()
    return int(rows), int(columns)
